#include "scan.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <complex.h>

#include "phy/signal.h"
#include "utils/file.h"
#include "phy/frequency_offset.h"
#include "utils/fft.h"
#include "phy/channel.h"
#include "phy/qpsk.h"
#include "transport/fic.h"
#include "transport/fig.h"
#include "coding/frequency_interleaving.h"
#include "transport/msc.h"
#include "sdr/sdr.h"
#include "utils/io.h"
#include "audio/alsa.h"
#include "utils/log.h"

static int time_sync(io *in)
{
  complex ts[T_F + T_NULL];
  int c;
  double energy;
  double min_energy;
  int min_pos;
  int i;

  c = in->get_samples(in, ts, T_F + T_NULL);
  if (c != T_F + T_NULL)
    return -1;

  energy = 0;

  for (i = 0; i < T_NULL; i++)
    energy += cabs(ts[i]);

  min_pos = 0;
  min_energy = energy;

  for (i = T_NULL; i < T_F + T_NULL; i++) {
    energy -= cabs(ts[i - T_NULL]);
    energy += cabs(ts[i]);
    if (energy < min_energy) {
      min_energy = energy;
      min_pos = i - T_NULL + 1;
    }
  }

  if (min_pos > T_F)
    min_pos -= T_F;

  in->put_back_samples(in, &ts[min_pos], T_F + T_NULL - min_pos);

  return min_pos;
}

static int fast_time_sync(io *in)
{
  complex ts[T_NULL + 200];
  int c;
  double energy;
  double min_energy;
  int min_pos;
  int i;

  c = in->get_samples(in, ts, T_NULL + 200);
  if (c != T_NULL + 200)
    return -1;

  energy = 0;

  for (i = 0; i < T_NULL; i++)
    energy += cabs(ts[i]);

  min_pos = 0;
  min_energy = energy;

  for (i = T_NULL; i < T_NULL + 200; i++) {
    energy -= cabs(ts[i - T_NULL]);
    energy += cabs(ts[i]);
    if (energy < min_energy) {
      min_energy = energy;
      min_pos = i - T_NULL + 1;
    }
  }

  if (min_pos > 200)
    min_pos -= 200;

  in->put_back_samples(in, &ts[min_pos], T_NULL + 200 - min_pos);

  return min_pos;
}

static struct {
  char *name;
  double freq;
} band3[] = {
{(char*)"5A", 174.928},
{(char*)"5B", 176.640},
{(char*)"5C", 178.352},
{(char*)"5D", 180.064},
{(char*)"6A", 181.936},
{(char*)"6B", 183.648},
{(char*)"6C", 185.360},
{(char*)"6D", 187.072},
{(char*)"7A", 188.928},
{(char*)"7B", 190.640},
{(char*)"7C", 192.352},
{(char*)"7D", 194.064},
{(char*)"8A", 195.936},     /* busy out3 */  /* busy out1 */ /* busy out2 (too loud) */
{(char*)"8B", 197.648},     /* busy out2 */
{(char*)"8C", 199.360},
{(char*)"8D", 201.072},     /* busy out3 */ /* busy out2 (loud) */
{(char*)"9A", 202.928},
{(char*)"9B", 204.640},
{(char*)"9C", 206.352},
{(char*)"9D", 208.064},     /* busy? out2 */
{(char*)"10A", 209.936},
{(char*)"10B", 211.648},
{(char*)"10C", 213.360},
{(char*)"10D", 215.072},
{(char*)"10N", 210.096},
{(char*)"11A", 216.928},    /* busy out3 */ /* busy out1 */ /* busy out2 (loud) */
{(char*)"11B", 218.640},    /* busy out2 */
{(char*)"11C", 220.352},    /* busy out3 */ /* busy out1 */ /* busy out2 */
{(char*)"11D", 222.064},
{(char*)"11N", 217.088},    /* busy out3 */ /* busy out1 */ /* busy out2 */
{(char*)"12A", 223.936},
{(char*)"12B", 225.648},    /* busy out3 */ /* busy out1 */ /* busy out2 */
{(char*)"12C", 227.360},    /* busy out3 (low) */ /* busy out2 */
{(char*)"12D", 229.072},
{(char*)"12N", 224.096},    /* busy out2 */
{(char*)"13A", 230.784},
{(char*)"13B", 232.496},
{(char*)"13C", 234.208},
{(char*)"13D", 235.776},
{(char*)"13E", 237.488},
{(char*)"13F", 239.200},
{NULL, 0}};

void scan(void)
{
  io *in;
  int i;
  int j;
  int start;
  double fine_freq_offset;
  complex *in_time;
  complex *_cur_freq;
  complex *_prev_freq;
  complex *cur_freq;
  complex *prev_freq;
  complex *channel;
  complex *z_ref;
  fft_data *fft_data;
  int freq_offset_cur_pos;
  complex qpsk[K];
  int p[K*2];
  fic *fic;
  frequency_interleaver *fl;
  frame_figs figs;
  int frame_cif_id;
  int fig0_0_errors;
  int fig0_0_present;
  int fig0_1_looped;
  int fig0_2_looped;
  int fig1_1_looped;
  fig0_1 f01[256];
  int f01_size;
  fig0_2 f02[256];
  int f02_size;
  fig1_1 f11[256];
  int f11_size;

  double start_frequency;

  fic = new_fic();

  fl = new_frequency_interleaver();

  init_frequency_offset();

  in_time = malloc(sizeof(complex) * T_S);
  _cur_freq = malloc(sizeof(complex) * T_U);
  _prev_freq = malloc(sizeof(complex) * T_U);
  channel = malloc(sizeof(complex) * T_U);
  z_ref = malloc(sizeof(complex) * T_U);
  if (in_time == NULL || _cur_freq == NULL || _prev_freq == NULL
      || channel == NULL || z_ref == NULL) abort();

  cur_freq = _cur_freq;
  prev_freq = _prev_freq;

  fft_data = init_fft(T_U);

  compute_phase_ref_full(fft_data, z_ref);

  in = new_sdr("usrp");
  if (in == NULL) exit(1);

  int fff = -1;

  in->set_gain(in, 1);
  in->set_freq(in, 10000000); /* dumb value (10MHz) for start */
  in->start(in);

next_freq:
  fff++;
  if (band3[fff].name == NULL) exit(0);
  printf("digging %s (%g)\n", band3[fff].name, band3[fff].freq);
  start_frequency = band3[fff].freq * 1000000;
  fig0_0_present = 0;
//start_frequency = 225648000;
printf("freq %g %g\n", start_frequency, band3[fff].freq * 1000000);
  in->reset(in);
  in->set_freq(in, start_frequency); //225648000); //220352000); //201072000); //195936000); //216928000);

restart:
printf("RESET LOOOPPP\n");
  fig0_1_looped = 0;
  fig0_2_looped = 0;
  fig1_1_looped = 0;
  fig0_0_errors = 0;
  f01_size = f02_size = f11_size = 0;

  start = time_sync(in);
  if (start == -1) { printf("no time sync\n"); goto next_freq; }

next_frame:
  debug("#start %d\n", start);

  if (in->get_samples(in, NULL, T_NULL) != T_NULL) goto over;

  figs.size = 0;

  /* symbol 1 */
  if (in->get_samples(in, in_time, T_S) != T_S) goto over;
  fine_freq_offset = frequency_offset_fine(in_time);
  debug("# %g ", fine_freq_offset);
  freq_offset_cur_pos = remove_frequency_offset(0, 
                            in_time, fine_freq_offset);
  fine_freq_offset = frequency_offset_fine(in_time);
  debug("%g\n", fine_freq_offset);

{
int i;
double e = 0;
for (i= 0; i < T_S; i++) e += cabs(in_time[i]);
printf("energy %g\n", e / T_S);
}

  fft(fft_data, &in_time[Delta], prev_freq);

  for (i = 2; i < 5; i++) {
    if (in->get_samples(in, in_time, T_S) != T_S) goto over;
    fine_freq_offset = frequency_offset_fine(in_time);
    debug("#%g ", fine_freq_offset);
    freq_offset_cur_pos = remove_frequency_offset(freq_offset_cur_pos,
                              in_time, fine_freq_offset);
    fine_freq_offset = frequency_offset_fine(in_time);
    debug("%g\n", fine_freq_offset);

    fft(fft_data, &in_time[Delta], cur_freq);

    //channel_compensate(cur_freq, channel, cur_freq);

    /* frequency de-interleaving done here (using k2n) */
    /* positive frequencies [1..768] */
    for (j = 0; j < K/2; j++)
      qpsk[fl->k2n[FK(j+1)]] = cur_freq[j+1] * conj(prev_freq[j+1]);
    /* negative frequencies [-768..-1] */
    for (j = 0; j < K/2; j++)
      qpsk[fl->k2n[FK(j-K/2)]] = cur_freq[j+T_U-K/2] * conj(prev_freq[j+T_U-K/2]);

    symbol_unmap(p, qpsk);

    /* symbols 2, 3 and 4 are for fic */
    if (i < 5) {
      for (j = 0; j < 4; j++) {
        /* FIB = 256 bits, 30 bytes of data, 2 bytes of crc
         * 4 FIBs per symbol, 12 FIBs for the 3 symbols of the FIC
         * 3 FIBs make for 1 FIC and are used for energy dispersal (768 bits)
         * FIC (768 bits) is convolution encoded, giving 3072+24 bits
         * 3072+24 bits are punctured, giving 2304 bits
         * puncturing is:
         *     21 blocks of 128 bits are punctured by PI=16 (rate 8/24)
         *     (we remove 8 bits over 32)
         *     3 blocks of 128 bits are punctured by PI=15 (rate 9/24)
         *     (we remove 9 bits over 32)
         *     last 24 bits are punctured by 1100 1100 1100 1100 .. 1100
         *     (we remove 12 bits over 24)
         * each symbol gets 3072 bits, so:
         * 1st symbol: fic 0 (2304 bits) + fic 1 (768 bits)
         * 2nd symbol: fic 1 (1536 bits) + fic 2 (1536 bits)
         * 3rd symbol: fic 2 (768 bits)  + fic 3 (2304 bits)
         * we send to the fic unit blocks of 768 bits
         * when the fic unit has enough bits it does the decoding
         */
        receive_fic_bits(&figs, fic, &p[768*j], (i-2)*4+j);
      }
    }

    if (i == 4) {
      int f;
      for (f = 0; f < figs.size; f++) {
        switch (figs.f[f].type) {
        default: break;
        case FIG0_1:
printf("got fig 0/1 loop %d size %d\n", fig0_1_looped, f01_size);
          if (fig0_1_looped) break;
          if (f01_size && !memcmp(&figs.f[f].u.fig0_1, &f01[0], sizeof(fig0_1))) {
            fig0_1_looped = 1;
            break;
          }
          if (f01_size == 256) abort();
          f01[f01_size] = figs.f[f].u.fig0_1;
          f01_size++;
          break;
        case FIG0_2:
printf("got fig 0/2 loop %d size %d\n", fig0_2_looped, f02_size);
          if (fig0_2_looped) break;
          if (f02_size && !memcmp(&figs.f[f].u.fig0_2, &f02[0], sizeof(fig0_2))) {
            fig0_2_looped = 1;
            break;
          }
          if (f02_size == 256) abort();
          f02[f02_size] = figs.f[f].u.fig0_2;
          f02_size++;
          break;
        case FIG1_1:
printf("got fig 1/1 loop %d size %d\n", fig1_1_looped, f11_size);
          if (fig1_1_looped) break;
          if (f11_size && !memcmp(&figs.f[f].u.fig0_1, &f11[0], sizeof(fig1_1))) {
            fig1_1_looped = 1;
            break;
          }
          if (f11_size == 256) abort();
          f11[f11_size] = figs.f[f].u.fig1_1;
          f11_size++;
          break;
        }
      }

      /* check that we have FIG 0/0 */
      if (figs.size) printf("fig\n");
      if (figs.size == 0 || figs.f[0].type != FIG0_0) {
printf("type %d size %d\n", figs.f[0].type, figs.size);
        debug("error: FIG 0/0 not received, skip frame\n");
        fig0_0_errors++;
        if (fig0_0_errors == 10) {
          debug("error: too much missed FIG 0/0, full resync\n");
          if (fig0_0_present) goto restart;
          goto next_freq;
        }
        i++;
        goto frame_over;
      }
      fig0_0_present++;
      if (fig0_1_looped && fig0_2_looped && fig1_1_looped) {
        goto next_freq;
      }
      fig0_0_errors = 0;
      frame_cif_id = figs.f[0].u.fig0_0.cif_count;
      i++;
      goto frame_over;
    }

#define  SWAP(a, b) do { complex *x = a; a = b; b = x; } while (0)
    SWAP(cur_freq, prev_freq);
  }

frame_over:
  /* skip msc symbols */
printf("i %d\n", i);
  for (; i < 77; i++) {
    if (in->get_samples(in, in_time, T_S) != T_S) goto over;
  }

  /* put back last 100 samples of last symbol, for fast time sync */
  in->put_back_samples(in, &in_time[T_S-100], 100);
  start = fast_time_sync(in);
  if (start == -1) goto restart;

  goto next_frame;

over:
  in->stop(in);
  in->free(in);
}
